//
//  View.swift
//  USB Meter
//
//  Created by Bogdan Timofte on 04/05/2020.
//  Copyright © 2020 Bogdan Timofte. All rights reserved.
//

import SwiftUI

/* MARK: Iusless...
enum XNavigationViewStyle {
    case auto
    case doubleColumn
    case stack
}

extension View {
    func xNavigationViewStyle(_ style: XNavigationViewStyle) -> some View {
        switch style {
        case .auto:
            track("auto")
            return AnyView(self.navigationViewStyle(DefaultNavigationViewStyle()))
        case .doubleColumn:
            track("doubleColumn")
            return AnyView(self.navigationViewStyle(DoubleColumnNavigationViewStyle()))
        case .stack:
            track("stack")
            return AnyView(self.navigationViewStyle(StackNavigationViewStyle()))
        }
    }
}
*/
