//
//  Path.swift
//  USB Meter
//
//  Created by Bogdan Timofte on 14/05/2020.
//  Copyright © 2020 Bogdan Timofte. All rights reserved.
//

import CoreGraphics
import SwiftUI

extension Path {
    mutating func addLine(from startPoint: CGPoint, to endPoint: CGPoint) {
        self.move(to: startPoint)
        self.addLine(to: endPoint)
    }
}
